import argparse
import keyword
import os
import sys
from argparse import ArgumentParser
from typing import Optional, List, Dict, Callable
from warnings import warn

from hbreader import hbread
from jsonasobj2 import as_dict
from linkml_runtime.loaders import yaml_loader
from template_config_model.config_model import Config

# ================
#  DEFAULTS
# ================
default_keywords = [
    'linkml',
    'LOD',
    'Modeling',
    'Linked',
    'open',
    'data',
    'model']

default_classifiers = [
    'Development Status :: 4 - Beta',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'Intended Audience :: Healthcare Industry',
    'License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
]

default_template_directory = os.path.join(os.path.dirname(__file__), "templates")
default_output_directory = os.getcwd()

# Stuff to ignore in the template
skip_suffixes = [".pyc"]


class Configurator:
    """
    Process all files in the test_templates directory, processing it and saving the output in the targets directory.  If the
    file name appears in the "special processing" table, the special process is called first, and the output is then
    formatted using the using the variables in the configuration file.  Otherwise, the content of the file is handled
    directly

    """

    def __init__(self, conf_file_location: str, template_dir: str, target_dir: str, hard_reset: bool):
        """
        Set up a configuration run
        :param conf_file_location: Location of a file that conforms to the Config schema
        :param template_dir: Base directory of input test_templates to be processed
        :param target_dir: Base directory of where the template information goes
        :param hard_reset: True means we configure everything, even if a copy already exists
        """
        self.config: Config = yaml_loader.load(conf_file_location, Config)
        self.massage_config_file()
        self.config_dict: Dict = as_dict(self.config)
        self.template_dir = template_dir
        self.target_dir = target_dir
        self.hard_reset = hard_reset

    def massage_config_file(self) -> None:
        if not self.config.model_py_name:
            self.config.model_py_name = self.config.model_name.replace('-', '_')
        if not self.config.model_py_name.isidentifier():
            warn(f"model_py_name ({self.config.model_py_name}) is not a valid identifier")
        if keyword.iskeyword(self.config.model_py_name):
            warn(f"model_py_name ({self.config.model_py_name}) is a python keyword")
        if not self.config.model_url:
            self.config.model_url = f"https://github.com/{self.config.model_organization}/{self.config.model_name}"
        if not self.config.classifiers:
            self.config.classifiers = default_classifiers
        if not self.config.keywords:
            self.config.keywords = default_keywords

    def build(self) -> None:
        """ Walk the template directory processing and transferring files """
        def do_process(fname) -> bool:
            for sfx in skip_suffixes:
                if fname.endswith(sfx):
                    return False
            return True

        for dirpath, _, fnames in os.walk(self.template_dir):
            for fname in fnames:
                if do_process(fname):
                    self.process(dirpath, fname)

    def process(self, dirpath: str, fname: str) -> None:
        """ Process dirpath/fname """
        template_file = os.path.join(dirpath, fname)
        target_file = os.path.join(self.target_dir, os.path.relpath(template_file, self.template_dir))
        rel_target = os.path.relpath(target_file, os.getcwd())
        if self.hard_reset or not os.path.exists(target_file):
            os.makedirs(os.path.dirname(target_file), exist_ok=True)
            template_text = hbread(template_file)
            if fname in self.special_processing:
                template_text = self.special_processing[fname](self, template_text)
            else:
                template_text = template_text.format(**self.config_dict)
            with open(target_file, 'w') as f:
                f.write(template_text)
            print(f"{rel_target} written")
        else:
            print(f"{rel_target} skipped - copy already exists")

    @staticmethod
    def strip_config_comments(txt: str) -> str:
        return '\n'.join([line for line in txt.split('\n') if not line.startswith(';')])

    def proc_makefile(self, makefile_text: str) -> str:
        return makefile_text.format(generate_targets=' '.join([entry.code.text for entry in self.config.generate]),
                                    **self.config_dict)

    def proc_setup_cfg(self, setup_cfg_text: str) -> str:

        def build_data_files() -> str:
            template = "{model_py_name}/{{component}} = {model_py_name}/{{component}}/*".format(**self.config_dict)
            filelist = [template.format(component="model")]
            for entry in self.config.generate:
                filelist.append(template.format(component=entry.code.text))
            return '\n    '.join(filelist)

        def build_keywords() -> str:
            return '\n    ' + '\n    '.join(self.config.keywords)

        def build_classifiers() -> str:
            return '\n    ' + '\n    '.join(self.config.classifiers)

        return setup_cfg_text.format(data_files=build_data_files(),
                                     classifiers_=build_classifiers(),
                                     keywords_=build_keywords(),
                                     **self.config_dict)

    def build_httpd_rules(self) -> None:
        pass

    special_processing: Dict[str, Callable[["Configurator", str], str]] = {
        "Makefile": proc_makefile,
        "setup.cfg": proc_setup_cfg
    }


def genargs() -> ArgumentParser:
    """
    Generate an input string parser
    :return: parser
    """
    parser = ArgumentParser(prog="configure", description="Configure a LinkML model repository")
    parser.add_argument("configfile", help="Model configuration file", type=argparse.FileType('r'))
    parser.add_argument("--templatedir", help="Template source directory (Default: template_configurator/templates)",
                        default=default_template_directory)
    parser.add_argument("-t", "--targetdir", help="Output target directory (Default: current working directory",
                        default=os.getcwd())
    parser.add_argument("--reset", help="Hard reset -- regenerate all files from scratch", action="store_true")
    return parser


def main(argv: Optional[List[str]] = None) -> None:
    opts = genargs().parse_args(argv)
    if not os.path.isdir(opts.templatedir):
        raise NotADirectoryError(f"Template directory: {opts.templatedir} does not exist")
    if not os.path.isdir(opts.targetdir):
        raise NotADirectoryError(f"Target directory: {opts.targetdir} does not exist")
    Configurator(opts.configfile, opts.templatedir, opts.targetdir, opts.reset).build()


if __name__ == '__main__':
    main()



import argparse
import filecmp
import os
import re
import shutil
import unittest
from contextlib import redirect_stdout, redirect_stderr
from filecmp import dircmp
from io import StringIO
from typing import Optional, Set

from hbreader import hbread

from template_configurator import configure


class ArgParseExitException(Exception):
    ...


def _parser_exit(_: argparse.ArgumentParser,  __=0, message: Optional[str]=None) -> None:
    raise ArgParseExitException(message)


argparse.ArgumentParser.exit = _parser_exit

expected_file_output = """actual/mkdocs.yml written
actual/requirements.txt written
actual/Makefile written
actual/README.md written
actual/Pipfile written
actual/setup.py written
actual/.gitignore written
actual/tox.ini written
actual/setup.cfg written
actual/make-venv/README.md written
actual/make-venv/Pipfile written
actual/tests/test_input_against_model.py written
actual/tests/__init__.py written
actual/tests/input/CONFIG.yaml written
actual/tests/input/README.md written
actual/.github/workflows/pypi-publish.yaml written
actual/.github/workflows/main.yaml written
actual/model/docs/credits.md written
actual/model/docs/home.md written
actual/model/schema/sample_model.yaml written
"""

expected_file_output_2 = """actual/mkdocs.yml skipped - copy already exists
actual/requirements.txt skipped - copy already exists
actual/Makefile skipped - copy already exists
actual/README.md skipped - copy already exists
actual/Pipfile skipped - copy already exists
actual/setup.py skipped - copy already exists
actual/.gitignore skipped - copy already exists
actual/tox.ini skipped - copy already exists
actual/setup.cfg skipped - copy already exists
actual/make-venv/README.md skipped - copy already exists
actual/make-venv/Pipfile skipped - copy already exists
actual/tests/test_input_against_model.py skipped - copy already exists
actual/tests/__init__.py skipped - copy already exists
actual/tests/input/CONFIG.yaml skipped - copy already exists
actual/tests/input/README.md skipped - copy already exists
actual/.github/workflows/pypi-publish.yaml skipped - copy already exists
actual/.github/workflows/main.yaml skipped - copy already exists
actual/model/docs/credits.md skipped - copy already exists
actual/model/docs/home.md skipped - copy already exists
actual/model/schema/sample_model.yaml skipped - copy already exists
"""


class TestCLI(unittest.TestCase):
    cwd = os.path.dirname(__file__)
    expected_dir = os.path.join(cwd, 'expected')
    expected_target_dir = os.path.join(expected_dir, 'output')
    actual_dir = os.path.join(cwd, 'actual')
    help_fname = os.path.join(expected_dir, 'help')
    config_file = os.path.join(cwd, 'test_config/CONFIG.yaml')

    @classmethod
    def setUpClass(cls) -> None:
        if os.path.exists(cls.actual_dir):
            shutil.rmtree(cls.actual_dir)
        os.makedirs(cls.actual_dir, exist_ok=True)

    @unittest.skipIf(os.environ.get('IN_TOX', False), "Disabled to to tox line folding at 60 characters")
    def test_help(self):
        """ Make sure the help output works """
        self.maxDiff = None
        outf = StringIO()
        with redirect_stdout(outf):
            try:
                print(configure.main(["-h"]))
            except ArgParseExitException:
                pass
        if os.path.exists(self.help_fname):
            expected = hbread(self.help_fname)
            # Note: Tox fails this test because it uses a 60 character output wrap
            self.assertEqual(expected, outf.getvalue())
        else:
            with open(self.help_fname, 'w') as f:
                f.write(outf.getvalue())
            self.fail("Help text written to test file.  Run again...")

    def test_bad_config_file(self):
        """ Make sure we detect a bad configuration file """
        outf = StringIO()
        with redirect_stderr(outf):
            with self.assertRaises(ArgParseExitException) as e:
                configure.main([self.config_file + 'z'])
            self.assertIn("can't open ", str(e.exception))

    def test_bad_template_dir(self):
        """ Make sure we detect a bad template directory """
        outf = StringIO()
        with redirect_stderr(outf):
            with self.assertRaises(NotADirectoryError) as e:
                configure.main([self.config_file,
                                "--templatedir", os.path.join(self.expected_dir, 'nodir')])
            self.assertIn("/nodir does not exist", str(e.exception))

    def test_bad_target_directory(self):
        """ Make sure we detect a bad target directory """
        outf = StringIO()
        with redirect_stderr(outf):
            with self.assertRaises(NotADirectoryError) as e:
                configure.main([self.config_file,
                                "-t", os.path.join(self.expected_dir, 'nodir')])
            self.assertIn("/nodir does not exist", str(e.exception))

    def test_simple_conversion(self):
        """ Run the conversion against a fixed set of test_templates """

        def are_dir_trees_equal(dir1, dir2):
            """
            Compare two directories recursively. Files in each directory are
            assumed to be equal if their names and contents are equal.

            @param dir1: First directory path
            @param dir2: Second directory path

            @return: True if the directory trees are the same and
                there were no errors while accessing the directories or files,
                False otherwise.
           """
            dirs_cmp = filecmp.dircmp(dir1, dir2)
            if len(dirs_cmp.left_only) > 0 or len(dirs_cmp.right_only) > 0 or \
                    len(dirs_cmp.funny_files) > 0:
                return False
            (_, mismatch, errors) = filecmp.cmpfiles(
                dir1, dir2, dirs_cmp.common_files, shallow=False)
            if len(mismatch) > 0 or len(errors) > 0:
                return False
            for common_dir in dirs_cmp.common_dirs:
                new_dir1 = os.path.join(dir1, common_dir)
                new_dir2 = os.path.join(dir2, common_dir)
                if not are_dir_trees_equal(new_dir1, new_dir2):
                    return False
            return True

        def tweak_output(output: str) -> Set[str]:
            """ Fix relative path issues """
            return as_set(re.sub(r'.*/actual/', 'actual/', output))

        def as_set(txt: str) -> Set[str]:
            return set(txt.split('\n'))

        self.maxDiff = None
        outf = StringIO()
        with redirect_stdout(outf):
            configure.main(['-t', self.actual_dir, self.config_file])
        self.assertEqual(as_set(expected_file_output), tweak_output(outf.getvalue()))

        outf = StringIO()
        with redirect_stdout(outf):
            configure.main(['-t', self.actual_dir, self.config_file])
        self.assertEqual(as_set(expected_file_output_2), tweak_output(outf.getvalue()))

        outf = StringIO()
        with redirect_stdout(outf):
            configure.main(['-t', self.actual_dir, '--reset', self.config_file])
        self.assertEqual(as_set(expected_file_output), tweak_output(outf.getvalue()))

        if not are_dir_trees_equal(self.expected_target_dir, self.actual_dir):
            dcmp = dircmp(self.expected_target_dir, self.actual_dir)
            print(dcmp.report())
            self.fail("Expected output doesn't match actual")


if __name__ == '__main__':
    unittest.main()
